''*************************************
''*  Virtual Graphics Driver v1.1     *
''*************************************
''
''By RossH
''
''REVISION HISTORY
''
'' Rev  Date      Description
'' ---  ---------  -----------------------------------------------------
'' 010   7 Jul 08  First version, based on Parallax Graphics Driver v1.0
''                 Supports 2 or 4 colour mode. For use with Virtual VGA Driver.
'' 011  28 Jul 08  Added support for reclaiming RAM
''
'' Triangle adaptation May 2010 JSD (Doesn't work)

CON

  #1, _setup, _color, _width, _plot, _line, _arc, _vec, _vecarc, _pix, _pixarc, _text, _textarc, _fill, _flush, _loop

  TILE_BITS     = 12            ' virtual VGA uses 12 bits of tile number (instead of the normal 10)
  TILE_MASK     = $0FFF         ' 12 bits of tile number allows for 4096 tiles, necessary for high resolutions
  COLOUR_MASK   = $F000         ' virtual VGA uses 4 bits of colour (instead of the normal 6) for 16 colour combinations

VAR

  long  cogon, cog
  long  command

' RJH - 2 or 4 color mode ------------------------------------------------------------------NEW
  word  colour_mode                                     '0 for 2 color mode, <> 0 for 4 colour mode
' RJH - ------------------------------------------------------------------------------------END
  word  tpm_base                                        'base of screen tile pointer map
  word  map_base                                        'base of tile translation map
  word  tpm_tiles                                       'number of tiles in screen tile pointer map
  long  tile_base                                       'base of screen tile buffer (aligned on 64 long boundary)
  long  tile_0                                          'address of all 0 tile
  long  tile_list                                       'pointer to list of free tiles
  long  tile_count                                      'count of available tiles

  long  bases[50]

  long  pixel_width                                     'pixel data
  long  slices[8]





OBJ

  db     : "RJH_Virtual_Graphics_db_011.spin"                ' Virtual Graphics Double Buffer Handler

PUB start : okay

'' Start graphics driver - starts a cog
'' returns false if no cog available


  stop
  okay := cogon := (cog := cognew(@loop, @command)) > 0
  if okay
    okay := db.start

PUB stop

'' Stop graphics driver - frees a cog

  if cogon~
    cogstop(cog)

  command~


PUB setup(mode, x_tiles, y_tiles, x_origin, y_origin, tpm_ptr, map_ptr, n_tiles, tile_ptr)  | n, x, y, ar0, ar1, ar2, ar3, ar4, ar5, ar6, ar7, ar8, ar9

'' Set bitmap and tile parameters
''
''   mode           - 0 means 2 colour mode, <> 0 means 4 colour mode
''   x_tiles        - number of x tiles (tiles are 16x16 pixels in 4 color mode or 32x16 pixels in 2 colour mode)
''   y_tiles        - number of y tiles
''   x_origin       - relative-x center pixel
''   y_origin       - relative-y center pixel
''
''   tpm_ptr        - pointer to tile pointer map - an array of x_tiles * y_tiles words
''   map_ptr        - pointer to tile translation map - an array of x_tiles * y_tiles words
''   n_tiles        - the number of tiles in the tpm - number of tiles in tile space
''   tile_ptr       - tile space pointer - pointer to 16*n_tiles longs (16 longs per tile). Must be 64 byte aligned.
''
' RJH - 2 or 4 color mode ----------------------------------------------------------------- OLD
' (for old code see the original parallax graphics driver)
' RJH - 2 or 4 color mode ----------------------------------------------------------------- NEW

  colour_mode := mode
  if colour_mode == 0                                   '2 colour mode - map colours to 0 or 3
    colors[1] :=%%3333_3333_3333_3333
    colors[2] :=%%3333_3333_3333_3333
    colors[3] :=%%3333_3333_3333_3333
  else                                                  '4 colour mode
    colors[1] :=%%1111_1111_1111_1111
    colors[2] :=%%2222_2222_2222_2222
    colors[3] :=%%3333_3333_3333_3333

  tile_base := tile_ptr
  tile_0    := tile_ptr

  repeat n from 0 to 15
     long[tile_0][n] := $0000_0000                      'first tile is special 'all 0' tile

  tile_list := 0
  tile_count := n_tiles
  repeat n from 1 to tile_count - 1                     'put remaining tiles ...
     long[tile_base][n * 16] := tile_list               '... in a linked list ...
     tile_list := tile_base + (n * 64)                  '... of free tiles

  'init map (used to map virtual address to tile addresses)
  map_base := map_ptr
  repeat x from 0 to x_tiles - 1
    repeat y from 0 to y_tiles - 1
      word[map_base][x*y_tiles + y] := tpm_ptr + (y*x_tiles + x)<<1

  'init tpm (initially all entries point to special 'all 0' tile, and have colour 0)
  tpm_base := tpm_ptr
  tpm_tiles := (x_tiles * y_tiles)                      'number of words in tile pointer map
  repeat n from 0 to tpm_tiles - 1                      'init tile screen to all blanks (tile_0)
    word[tpm_base][n] := tile_0>>6

  setcommand(_loop, 0)                                  'make sure last command finished

  'for the virtual vga driver we have no actual bitmap, but for the purposes
  'of calculations we pretend we have one and it starts at zero. The map and bases are
  'used to translate these virtual addresses to physical tile addresses
  'Note: 50 bases is sufficient for 1600 (2 colour), or 800 (4 color) horizontal pixel resolution.
  repeat n from 0 to x_tiles - 1 <# 50                  'write bases
    bases[n] := n * y_tiles << 6

  y_tiles <<= 4                                         'adjust arguments and do setup command
  y_origin := y_tiles - y_origin - 1

  ar0 := mode
  ar1 := x_tiles
  ar2 := y_tiles
  ar3 := x_origin
  ar4 := y_origin
  ar5 := map_ptr
  ar6 := tile_0
  ar7 := @tile_list
  ar8 := @bases
  ar9 := @slices

  setcommand(_setup, @ar0)

  ' start double buffer assistance handler but with double buffering diabled
  ' db_setup must be called later to enable double buffering
  db.setup (tpm_base, 0, 0, tile_0, tpm_tiles, @tile_list)

' RJH - 2 or 4 color mode ----------------------------------------------------------------- END

PUB db_setup (tpm_dbuf, double_buffer)

'' Enable double buffering

  flush                                                'make sure last draw command flushed

  db.setup (tpm_base, tpm_dbuf, double_buffer, tile_0, tpm_tiles, @tile_list)

PUB clear

  flush                                                'make sure last draw command flushed

  db.clear

PUB copy(tpm_dbuf)

  flush                                                'make sure last draw command flushed

  db.copy(tpm_dbuf)

PUB move(tpm_dbuf)

  flush                                                'make sure last draw command flushed

  db.move(tpm_dbuf)

PUB addram(rstart, rend) | n, i

  flush                                                 'make sure last draw command flushed

  tile_base := (rstart + $003F) & $FFC0                 'align to 64 byte boundary
  n := (rend - tile_base)/64                            'number of tiles
  tile_count += n

  if n > 0
    bytefill(tile_base, 0, n*64)
    repeat i from 0 to n - 1                            'put additional tiles ...
      long[tile_base][i * 16] := tile_list              '... in linked list ...
      tile_list := tile_base + (i * 64)                 '... of free tiles

PUB color(c)

'' Set pixel color to two-bit pattern
''
''   c              - color code in bits[1..0]

  setcommand(_color, @colors[c & 3])                    'set color (in 2 color mode, colours 2 and 3 are the same as 1)


PUB width(w)  | pixel_passes, round_pix, i, p

'' Set pixel width
'' actual width is w[3..0] + 1
''
''   w              - 0..15 for round pixels, 16..31 for square pixels

  round_pix := not w & $10                              'determine pixel shape/width
  w &= $F
  pixel_width := w
  pixel_passes := w >> 1 + 1

  setcommand(_width, @w)                                'do width command now to avoid updating slices when busy

  p := (w ^ $F) & $F                                    'update slices to new shape/width
  repeat i from 0 to w >> 1
' RJH - 2 or 4 color mode ----------------------------------------------------------------- OLD
'    slices[i] := true >> (p << 1) << (p & $E)          '4 colour mode
' RJH - 2 or 4 color mode ------------------------------------------------------------------NEW
    if colour_mode <> 0
      slices[i] := true >> (p << 1) << (p & $E)         '4 colour mode
    else
      slices[i] := true >> (16+p) << ((16+p)>>1)        '2 colour mode
' RJH ------------------------------------------------------------------------------------- END
    if round_pix and pixels[w] & |< i
      p += 2
    if round_pix and i == pixel_passes - 2
      p += 2


PUB colorwidth(c, w)

'' Set pixel color and width

  color(c)
  width(w)


PUB plot(x, y)

'' Plot point
''
''   x,y            - point

  setcommand(_plot, @x)


PUB line(x, y)

'' Draw a line to point
''
''   x,y            - endpoint

  setcommand(_line, @x)


{
PUB box(x, y, box_width, box_height) | x2, y2, pmin, pmax

'' Draw a box with round/square corners, according to pixel width
''
''   x,y            - box left, box bottom

  if box_width > pixel_width and box_height > pixel_width

    pmax := pixel_width - (pmin := pixel_width >> 1)    'get pixel-half-min and pixel-half-max

    x += pmin                                           'adjust coordinates to accomodate width
    y += pmin
    x2 := x + box_width - 1 - pixel_width
    y2 := y + box_height - 1 - pixel_width

    plot(x, y)                                          'plot round/square corners
    plot(x, y2)
    plot(x2, y)
    plot(x2, y2)

    fill(x, y2 + pmax, 0, (x2 - x) << 16, 0, 0, pmax)   'fill gaps
    fill(x, y, 0, (x2 - x) << 16, 0, 0, pmin)
    fill(x - pmin, y2, 0, (x2 - x + pixel_width) << 16, 0, 0, y2 - y)


PUB quad(x1, y1, x2, y2, x3, y3, x4, y4)

'' Draw a solid quadrilateral
'' vertices must be ordered clockwise or counter-clockwise

  tri(x1, y1, x2, y2, x3, y3)                           'draw two triangle to make 4-sides polygon
  tri(x3, y3, x4, y4, x1, y1)

}
PUB tri(x1, y1, x2, y2, x3, y3)  | xy[2],  x, y, da, db1, db2, linechange, lines_minus_1

'' Draw a solid triangle

' reorder vertices by descending y

  case (y1 => y2) & %100 | (y2 => y3) & %010 | (y1 => y3) & %001
    %000:
      longmove(@xy, @x1, 2)
      longmove(@x1, @x3, 2)
      longmove(@x3, @xy, 2)
    %010:
      longmove(@xy, @x1, 2)
      longmove(@x1, @x2, 4)
      longmove(@x3, @xy, 2)
    %011:
      longmove(@xy, @x1, 2)
      longmove(@x1, @x2, 2)
      longmove(@x2, @xy, 2)
    %100:
      longmove(@xy, @x3, 2)
      longmove(@x2, @x1, 4)
      longmove(@x1, @xy, 2)
    %101:
      longmove(@xy, @x2, 2)
      longmove(@x2, @x3, 2)
      longmove(@x3, @xy, 2)

' draw triangle
'(vars seperated to work out the math by hand)
  x := x1
  y := y1
  da := (x3 - x1) << 16 / (y1 - y3 + 1)
  db1 := (x2 - x1) << 16 / (y1 - y2 + 1)
  db2 := (x3 - x2) << 16 / (y2 - y3 + 1)
  linechange := y1 - y2
  lines_minus_1 := y1 - y3

  fill(x, y, da, db1, db2, linechange, lines_minus_1)
  'fill(x1, y1, (x3 - x1) << 16 / (y1 - y3 + 1), (x2 - x1) << 16 / (y1 - y2 + 1), (x3 - x2) << 16 / (y2 - y3 + 1), y1 - y2, y1 - y3)


PUB finish

'' Wait for any current graphics command to flush
'' use this to insure that it is safe to manually manipulate the bitmap

  setcommand(_loop, 0)                                'make sure last command flushed

PUB flush

'' Flush any remembered addresses

  setcommand(_flush, 0)                                'make sure last command flushed


PRI fill(x, y, da, db1, db2, linechange, lines_minus_1)

  setcommand(_fill, @x)


PUB tiles
  return tile_count

PUB ramstart
  return @ram_start

PUB ramend
  return @ram_end

PUB db_ramstart
  return db.ramstart

PUB db_ramend
  return db.ramend



PRI setcommand(cmd, argptr)

  command := cmd << 16 + argptr                      'write command and pointer
  repeat while command                               'wait for command to be cleared, signifying receipt




DAT

' Color codes
colors  long    %%0000000000000000
        long    %%1111111111111111
        long    %%2222222222222222
        long    %%3333333333333333

' Round pixel recipes

pixels  byte    %00000000,%00000000,%00000000,%00000000         '0,1,2,3
        byte    %00000000,%00000000,%00000010,%00000101         '4,5,6,7
        byte    %00001010,%00001010,%00011010,%00011010         '8,9,A,B
        byte    %00110100,%00111010,%01110100,%01110100         'C,D,E,F

DAT

'*************************************
'* Assembly language graphics driver *
'*************************************

                        org
ram_start
'
'
' Graphics driver - main loop
'
loop                    rdlong  t1,par          wz      'wait for command
        if_z            jmp     #loop

                        movd    :arg,#arg0              'get 10 arguments
                        mov     t2,t1
                        mov     t3,#10
:arg                    rdlong  arg0,t2
                        add     :arg,d0
                        add     t2,#4
                        djnz    t3,#:arg

                        wrlong  zero,par                'zero command to signify received

                        call    #setd                   'set dx,dy from arg0,arg1

                        ror     t1,#16+2                'lookup command address
                        add     t1,#jumps
                        movs    :table,t1
                        rol     t1,#2
                        shl     t1,#3
:table                  mov     t2,0
                        shr     t2,t1
                        and     t2,#$FF
                        jmp     t2                      'jump to command


jumps                   byte    0                       '0
                        byte    setup_                  '1
                        byte    color_                  '2
                        byte    width_                  '3
                        byte    plot_                   '4
                        byte    line_                   '5
                        byte    0                    '6
                        byte    0                    '7
                        byte    0                 '8
                        byte    0                    '9
                        byte    0                 'A
                        byte    0                   'B
                        byte    0                'C
                        byte    fill_               'D
                        byte    flush_                  'E
                        byte    loop                    'F

'
'
' setup(x_tiles, y_tiles, x_origin, y_origin, tpm_ptr, map_ptr, n_tiles, tile_ptr)  | t0, t1, tlist_ptr, bases_ptr, slices_ptr, n
'

setup_                  movs    :load,#arg0             'move ...
                        movd    :load,#cmode
                        mov     t2,#10                  '... 10 setup parameters
:load                   mov     0,0
                        add     :load,#1
                        add     :load,d0
                        djnz    t2,#:load

flush_                  mov     prevtile,color1          'don't have a previous tile (color1 is invalid as tile address)
'                        mov     prevaddr,color1          'don't have a previous address

                        jmp     #loop
'
'
' color(c)
'
color_                  mov     pcolor,arg0             'set pixel color

                        jmp     #loop
'
'
' width(w)  pixel_passes
'
width_                  mov     pwidth,arg0             'set pixel width
                        mov     passes,arg1             'set pixel passes

                        jmp     #loop
'
'
' plot(x, y)
'
plot_                   call    #plotd

                        jmp     #loop
'
'
' line(x, y)
'
line_                   call    #linepd

                        jmp     #loop


' RJH - virtual update -------------------------------------------------------------------- NEW
'
' RJH: perform ` update instead of real update
'  old code typically looks like:
'                       rdlong  xxx,long_addr
'                       andn    xxx,mask
'                       or      xxx,bits
'                       wrlong  xxx,long_addr
'
'  new code typically looks like:
'                       mov     addr,long_addr
'                       movs    mask,#mask
'                       movs    bits,#bits
'                       call    rdwrsparse
'
'  pseudo code of replacement:
'    find tile that contains addr
'    if tile is dummy tile_0, replace with real tile
'    if no tile available, replace with tile_0
'
' TODO: TBD: Reclaim blank tiles during spare cycles !!! (make sure not to add tile_0 to free list, and also fix up prevtile and prevaddr)
'
rdwrsparse
'                        cmp     addr,prevaddr    wz     'same virtual addr as last time?
'        if_z            jmp     #:have_addr             'yes - still have real addr
'                        mov     prevaddr,addr           'no - save virtual addr for next time
                        mov     tmp,addr                'convert virtual address ...
                        shr     tmp,#6                  ' ... to tile number (recall that virtual bitmap starts at 0)
                        cmp     tmp,prevtile    wz      'same tile as last time?
        if_z            jmp     #:have_tile_1           'yes - already have the real tile address
                        mov     prevtile,tmp            'no - save tile for next time
                        mov     tpmaddr,tmp             'get ...
                        shl     tpmaddr,#1              '... real tile ...
                        add     tpmaddr,mapptr
                        rdword  tpmaddr,tpmaddr
                        rdword  tileaddr,tpmaddr        ' ... tpm ...
                        mov     tilecolor,tileaddr      ' ... separating ...
                        and     tileaddr,tmask          ' ... tile ...
                        andn    tilecolor,tmask         ' ... from color
                        shl     tileaddr,#6             'tileaddr is the real address of the tile
                        cmp     tileaddr,t0ptr  wz      'is this the special 'all 0' tile?
        if_nz           jmp     #:have_tile_2           'no - we already have a tile to write on
                        rdlong  tileaddr,tlistptr       'yes - get address of first tile on free list
                        tjz     tileaddr,#:use_tile_0   ' if no tiles left, use special 'all 0' tile
                        rdlong  tmp,tileaddr            'yes - take tile off ...
                        wrlong  tmp,tlistptr            ' ... free ...
                        wrlong  zero,tileaddr           ' ... list
                        jmp     #:save_tile

:use_tile_0             mov     tileaddr,t0ptr          'save
:save_tile              mov     tmp,tileaddr            'save tile ...
                        shr     tmp,#6                  ' .. in ...
                        or      tmp,tilecolor           ' ... tile ...
                        wrword  tmp,tpmaddr             ' ... pointer map

:have_tile_1            cmp     tileaddr,t0ptr  wz      'set z flag if the tile we have is the 'all 0' tile
:have_tile_2            mov     realaddr,addr           'calculate
                        and     realaddr,#$3f           ' ... real ...
                        add     realaddr,tileaddr       ' ... address

:have_addr    if_nz     rdlong  tmp,realaddr            ' write ...
mask          if_nz     andn    tmp,0                   ' ... if not ...
bits          if_nz     or      tmp,0                   ' ... the 'all 0' ...
              if_nz     wrlong  tmp,realaddr            ' tile

rdwrsparse_ret          ret
' RJH ------------------------------------------------------------------------------------- END


'
'
' fill(x, y, da, db, db2, linechange, lines_minus_1)
'

fill_                   shl     dx,#16                  'get left and right fractions
                        or      dx,h8000
                        mov     t1,dx

                        mov     t2,xlongs               'get x pixels
                        shl     t2,#4

                        add     arg6,#1                 'pre-increment line counter

:yloop                  add     dx,arg2                 'adjust left and right fractions
                        add     t1,arg3

                        cmps    dx,t1           wc      'get left and right integers
        if_c            mov     base0,dx
        if_c            mov     base1,t1
        if_nc           mov     base0,t1
        if_nc           mov     base1,dx
                        sar     base0,#16
                        sar     base1,#16

                        cmps    base0,t2        wc      'left out of range?
        if_c            cmps    hFFFFFFFF,base1 wc      'right out of range?
        if_c            cmp     dy,ylongs       wc      'y out of range?
        if_nc           jmp     #:skip                  'if any, skip

                        mins    base0,#0                'limit left and right
                        maxs    base1,t2        wc
        if_nc           sub     base1,#1

                        shl     base0,#1                'make left mask
                        neg     mask0,#1
                        shl     mask0,base0
                        shr     base0,#5

                        shl     base1,#1                'make right mask
                        xor     base1,#$1E
                        neg     mask1,#1
                        shr     mask1,base1
                        shr     base1,#5

                        sub     base1,base0     wz      'ready long count
                        add     base1,#1

        if_z            and     mask0,mask1             'if single long, merge masks

                        shl     base0,#1                'get long base
                        add     base0,basesptr
                        rdword  base0,base0
                        shl     dy,#2
                        add     base0,dy
                        shr     dy,#2

                        mov     bits0,mask0             'ready left mask
:xloop                  mov     bits1,pcolor            'make color mask
                        and     bits1,bits0
' RJH - virtual update -------------------------------------------------------------------- OLD
                        'rdlong  pass,base0              'read-modify-write long
                        'andn    pass,bits0
                        'or      pass,bits1
                        'wrlong  pass,base0
' RJH - virtual update -------------------------------------------------------------------- NEW
                        mov     addr,base0
                        movs    mask,#bits0
                        movs    bits,#bits1
                        call    #rdwrsparse
' RJH ------------------------------------------------------------------------------------- END
                        shl     ylongs,#2               'advance to next long
                        add     base0,ylongs
                        shr     ylongs,#2
                        cmp     base1,#2        wz      'one more?
        if_nz           neg     bits0,#1                'if not, ready full mask
        if_z            mov     bits0,mask1             'if one more, ready right mask
                        djnz    base1,#:xloop           'loop if more longs

:skip                   sub     arg5,#1         wc      'delta change?
        if_c            mov     arg3,arg4               'if so, set new deltas
:same
                        add     dy,#1                   'adjust y
                        djnz    arg6,#:yloop            'another y?

                        jmp     #loop

'
'
' Plot line from px,py to dx,dy
'
linepd                  cmps    dx,px           wc, wr  'get x difference
                        negc    sx,#1                   'set x direction

                        cmps    dy,py           wc, wr  'get y difference
                        negc    sy,#1                   'set y direction

                        abs     dx,dx                   'make differences absolute
                        abs     dy,dy

                        cmp     dx,dy           wc      'determine dominant axis
        if_nc           tjz     dx,#:last               'if both differences 0, plot single pixel
        if_nc           mov     count,dx                'set pixel count
        if_c            mov     count,dy
                        mov     ratio,count             'set initial ratio
                        shr     ratio,#1
        if_c            jmp     #:yloop                 'x or y dominant?


:xloop                  call    #plotp                  'dominant x line
                        add     px,sx
                        sub     ratio,dy        wc
        if_c            add     ratio,dx
        if_c            add     py,sy
                        djnz    count,#:xloop

                        jmp     #:last                  'plot last pixel


:yloop                  call    #plotp                  'dominant y line
                        add     py,sy
                        sub     ratio,dx        wc
        if_c            add     ratio,dy
        if_c            add     px,sx
                        djnz    count,#:yloop

:last                   call    #plotp                  'plot last pixel

linepd_ret              ret
'
'
' Plot pixel at px,py
'
plotd                   mov     px,dx                   'set px,py to dx,dy
                        mov     py,dy

plotp                   tjnz    pwidth,#wplot           'if width > 0, do wide plot

                        mov     t1,px                   'compute pixel mask

' RJH 2 or 4 colour mode ------------------------------------------------------------------ OLD
'                        shl     t1,#1
'                        mov     mask0,#%11
' RJH 2 or 4 colour mode ------------------------------------------------------------------ OLD
                        test    cmode,#1        wc
        if_nc           mov     mask0,#%1               ' 2 colour mode
        if_c            shl     t1,#1                   ' 4 colour mode
        if_c            mov     mask0,#%11              ' 4 colour mode

' RJH ------------------------------------------------------------------------------------- END
                        shl     mask0,t1
                        shr     t1,#5

                        cmp     t1,xlongs       wc      'if x or y out of bounds, exit
        if_c            cmp     py,ylongs       wc
        if_nc           jmp     #plotp_ret

                        mov     bits0,pcolor            'compute pixel bits
                        and     bits0,mask0

                        shl     t1,#2                   'get address of pixel long
                        add     t1,basesptr
                        mov     t2,py
                        rdlong  t1,t1
                        shl     t2,#2
                        add     t1,t2

' RJH - virtual update -------------------------------------------------------------------- OLD
                        'rdlong  t2,t1                   'write pixel
                        'andn    t2,mask0
                        'or      t2,bits0
                        'wrlong  t2,t1
' RJH - virtual update -------------------------------------------------------------------- NEW
                        mov     addr,t1                  'write pixel
                        movs    mask,#mask0
                        movs    bits,#bits0
                        call    #rdwrsparse
' RJH - virtual update -------------------------------------------------------------------- END

plotp_ret
plotd_ret               ret

setd                    mov     dx,xorigin              'set dx,dy from arg0,arg1
                        add     dx,arg0
                        mov     dy,yorigin
                        sub     dy,arg1
setd_ret                ret

'
'
' Plot wide pixel
'
wplot                   mov     t1,py                   'if y out of bounds, exit
                        add     t1,#7
                        mov     t2,ylongs
                        add     t2,#7+8
                        cmp     t1,t2           wc
        if_nc           jmp     #plotp_ret

                        mov     t1,px                   'determine x long pair
' RJH - 4 color mode ---------------------------------------------------------------------- OLD
'                        sub     t1,#8
'                        sar     t1,#4
' RJH - 2 or 4 color mode ----------------------------------------------------------------- NEW
                        test    cmode,#1        wc
        if_nc           sub     t1,#16                  '2 colour mode
        if_nc           sar     t1,#5
        if_c            sub     t1,#8                   '4 colour mode
        if_c            sar     t1,#4
' RJH ------------------------------------------------------------------------------------- END
                        cmp     t1,xlongs       wc
                        muxc    jumps,#%01              '(use jumps[1..0] to store writes)
                        add     t1,#1
                        cmp     t1,xlongs       wc
                        muxc    jumps,#%10

                        test    jumps,#%11      wz      'if x out of bounds, exit
        if_z            jmp     #plotp_ret

                        shl     t1,#2                   'get base pair
                        add     t1,basesptr
                        rdlong  base1,t1
                        sub     t1,#4
                        rdlong  base0,t1

                        mov     t1,px                   'determine pair shifts
' RJH - 4 color mode ---------------------------------------------------------------------- OLD
'                        shl     t1,#1
'                        movs    :shift1,t1
'                        xor     :shift1,#7<<1
'                        add     t1,#9<<1
'                        movs    :shift0,t1
'                        test    t1,#$F<<1       wz      '(account for special case)
' RJH - 2 or 4 color mode ----------------------------------------------------------------- NEW
                        tjnz    cmode,#:wplot_1         ' 4 color mode
                        movs    :shift1,t1             ' 2 colour mode
                        xor     :shift1,#15
                        add     t1,#16
                        movs    :shift0,t1
                        test    t1,#$1F        wz
                        jmp     #:wplot_2
:wplot_1
                        shl     t1,#1                   '4 colour mode
                        movs    :shift1,t1
                        xor     :shift1,#7<<1
                        add     t1,#9<<1
                        movs    :shift0,t1
                        test    t1,#$F<<1       wz      '(account for special case)
:wplot_2
' RJH ------------------------------------------------------------------------------------- END
        if_z            andn    jumps,#%01

                        mov     pass,#0                 'ready to plot slices
                        mov     slice,slicesptr

:loop                   rdlong  mask0,slice             'get next slice
                        mov     mask1,mask0

:shift0                 shl     mask0,#0                'position slice
:shift1                 shr     mask1,#0

                        mov     bits0,pcolor            'colorize slice
                        and     bits0,mask0
                        mov     bits1,pcolor
                        and     bits1,mask1

                        mov     t1,py                   'plot lower slice
                        add     t1,pass
                        cmp     t1,ylongs       wc
        if_c            call    #wslice

                        mov     t1,py                   'plot upper slice
                        test    pwidth,#1       wc
                        subx    t1,pass
                        cmp     t1,ylongs       wc
        if_c            call    #wslice

                        add     slice,#4                'next slice
                        add     pass,#1
                        cmp     pass,passes     wz
        if_nz           jmp     #:loop

                        jmp     #plotp_ret
'
'
' Plot wide pixel slice
'
wslice                  shl     t1,#2                   'ready long offset

                        add     base0,t1                'plot left slice
                        test    jumps,#%01      wc

' RJH - virtual update -------------------------------------------------------------------- OLD
'       if_c            rdlong  t2,base0
'       if_c            andn    t2,mask0
'       if_c            or      t2,bits0
'       if_c            wrlong  t2,base0
' RJH ------------------------------------------------------------------------------------- END
        if_c            mov     addr,base0
        if_c            movs    mask,#mask0
        if_c            movs    bits,#bits0
        if_c            call    #rdwrsparse
' RJH - virtual update -------------------------------------------------------------------- END
                        add     base1,t1                'plot right slice
                        test    jumps,#%10      wc
' RJH - virtual update -------------------------------------------------------------------- OLD
'       if_c            rdlong  t2,base1
'       if_c            andn    t2,mask1
'       if_c            or      t2,bits1
'       if_c            wrlong  t2,base1
' RJH - virtual update -------------------------------------------------------------------- NEW
        if_c            mov     addr,base1
        if_c            movs    mask,#mask1
        if_c            movs    bits,#bits1
        if_c            call    #rdwrsparse
' RJH ------------------------------------------------------------------------------------- END

                        sub     base0,t1                'restore bases
                        sub     base1,t1

wslice_ret              ret

'
'
' Multiply
'
'   in:         t1 = 16-bit multiplicand (t1[31..16] must be 0)
'               t2 = 16-bit multiplier
'
'   out:        t1 = 32-bit product
'
multiply                mov     t3,#16
                        shl     t2,#16
                        shr     t1,#1           wc

:loop   if_c            add     t1,t2           wc
                        rcr     t1,#1           wc
                        djnz    t3,#:loop

multiply_ret            ret
'
'
' Defined data
'
zero                    long    0                       'constants
d0                      long    $200
h8000                   long    $8000
hFFFFFFFF               long    $FFFFFFFF
color1                  long    %%1111111111111111
color2                  long    %%2222222222222222
tmask                   long    TILE_MASK

fontptr                 long    0                       'font pointer (set before cognew command)

pcolor                  long    %%1111111111111111      'pixel color
pwidth                  long    0                       'pixel width
passes                  long    1                       'pixel passes
textsx                  long    1                       'text scale x
textsy                  long    1                       'text scale y
textsp                  long    6                       'text spacing
'
'
' Undefined data
'
t1                      res     1       'temps
t2                      res     1
t3                      res     1
t4                      res     1
t5                      res     1
t6                      res     1
t7                      res     1

arg0                    res     1       'arguments passed from high-level
arg1                    res     1
arg2                    res     1
arg3                    res     1
arg4                    res     1
arg5                    res     1
arg6                    res     1
arg7                    res     1
arg8                    res     1
arg9                    res     1

cmode                   res     1
xlongs                  res     1       'bitmap metrics
ylongs                  res     1
xorigin                 res     1
yorigin                 res     1
mapptr                  res     1       'address of translation map
t0ptr                   res     1       'address of the special 'all 0' tile
tlistptr                res     1       'address of free tile list ptr
basesptr                res     1       'pointers
slicesptr               res     1

dx                      res     1       'line/plot coordinates
dy                      res     1
px                      res     1
py                      res     1

sx                      res     1       'line
sy                      res     1
count                   res     1
ratio                   res     1

pass                    res     1       'plot
slice                   res     1
base0                   res     1
base1                   res     1
mask0                   res     1
mask1                   res     1
bits0                   res     1
bits1                   res     1

' RJH - virtual driver -------------------------------------------------------------------- NEW
addr                    res     1
'prevaddr                res     1
prevtile                res     1
realaddr                res     1
tileaddr                res     1
tilecolor               res     1
tpmaddr                 res     1
tmp                     res     1
' RJH ------------------------------------------------------------------------------------- END

ram_end
